import string
import sre_parse, Expression

class MultigroupPattern:
    # master pattern object.  keeps track of global attributes
    def __init__(self):
        self.flags = 0
        self.groups = 1
        self.groupdict = {}
    def reverse_name(self, id):
        for key, val in self.groupdict.items():
            if id in val:
                return key
        # Ignore non-named groups

    def getgroup(self, name=None):
        gid = self.groups
        self.groups = gid + 1
        if name:
            self.groupdict[name] = self.groupdict.get(name, []) + [gid]
        return gid


def convert_literal(pattern, name, val):
    return Expression.Literal(chr(val), 0)

def convert_not_literal(pattern, name, val):
    return Expression.Literal(chr(val), 1)

def convert_at(pattern, name, where):
    if where == "at_beginning":
        return Expression.AtBeginning()
    elif where == "at_end":
        return Expression.AtEnd()
    raise AssertionError("Unknown at name: %s" % repr(where))

def convert_any(pattern, name, ignore):
    assert ignore is None, "what does it mean when the field is '%s'?" % ignore
    return Expression.Any()

def convert_assert(pattern, name, (direction, terms)):
    assert direction == 1, "does not support lookbehind"
    return Expression.Assert(convert_list(pattern, terms), 0)

def convert_assert_not(pattern, name, (direction, terms)):
    assert direction == 1, "does not support lookbehind"
    return Expression.Assert(convert_list(pattern, terms), 1)


def convert_branch(pattern, name, (ignore, branches)):
    assert ignore is None, "what is %s?" % repr(ignore)
    results = []
    for branch in branches:
        results.append(convert_list(pattern, branch))
    if len(results) == 1:
        return results[0]
    return Expression.Alt(tuple(results))

# I know, it's only good for ASCII...
def invert(s):
    letters = []
    for c in map(chr, range(256)):
        if c not in s:
            letters.append(c)
    return string.join(letters, "") 

categories = {
    "category_word": string.letters + "0123456789_",
    "category_digit": string.digits,
    "category_space": "\t\n\v\f\r ",
 
    "category_not_word": invert(string.letters + "0123456789_"),
    "category_not_digit": invert(string.digits),
    "category_not_space": invert("\t\n\v\f\r "),
    }

def convert_in(pattern, name, terms):
    negate = (terms[0][0] == 'negate')
    s = ""
    for c in terms[negate:]:
        if c[0] == 'literal':
            s = s + chr(c[1])
        elif c[0] == 'range':
            for i in range(c[1][0], c[1][1]+1):
                s = s + chr(i)
        elif c[0] == 'category':
            s = s + categories[c[1]]
        else:
            raise AssertionError("unknown option for 'in': %s" % c[0])
    return Expression.In(s, negate)

def convert_subpattern(pattern, name, (id, terms)):
    pattern_name = pattern.reverse_name(id)
    return Expression.Subpattern(pattern_name, convert_list(pattern, terms))

def convert_max_repeat(pattern, name, (min_count, max_count, terms)):
    return Expression.MaxRepeat(convert_list(pattern, terms),
                                min_count, max_count)

def convert_groupref(pattern, name, id):
    assert type(id) != type(0), \
           "Martel cannot use numbered group reference: %d" % id
    # sre_parse returns the list from the MultigroupPattern
    # Map that back to a number.
    pattern_name = pattern.reverse_name(id[0])
    return Expression.GroupRef(pattern_name)

converter_table = {
    "literal": convert_literal,
    "not_literal": convert_not_literal,
    "branch": convert_branch,
    "in": convert_in,
    "subpattern":  convert_subpattern,
    "max_repeat":  convert_max_repeat,
    "at": convert_at,
    "any": convert_any,
    "assert": convert_assert,
    "assert_not": convert_assert_not,
    "groupref": convert_groupref,
    }

def convert_list(pattern, terms):
    # This is always a sequence of terms
    results = []
    for term in terms:
        name = term[0]
        try:
            func = converter_table[name]
        except KeyError:
            raise AssertionError, "Do not understand sre expression %s" % \
                  repr(name)
                  
        results.append( apply(func, (pattern,) + term) )
    if len(results) == 1:
        return results[0]
    return Expression.Seq(tuple(results))
        
    

def make_expression(s):
    x = sre_parse.parse(s, pattern = MultigroupPattern())
    return convert_list(x.pattern, x)


def _test():
    data = (
        ("a", "a"),
        ("ab", "ab"),
        ("a|b", "[ab]"),
        ("[ab]", "[ab]"),
        ("a|b|c|d", "[a-d]"),
        ("ab|bc", "ab|bc"),
        ("a*", "a*"),
        ("a+", "a+"),
        ("a?", "a?"),
        ("a{3}", "a{3}"),
        ("a{3,8}", "a{3,8}"),
        ("a{3,3}", "a{3}"),
        ("(?:test1)", "(test1)"),
        ("(?P<foo>test1)", "(?P<foo>test1)"),
        (".", "."),
        ("((?P<x>..?)|(?P<y>..*))", "((?P<x>..?)|(?P<y>..*))"),
        ("\w", "[\\dA-Z_a-z]"),
        ("\s", "[\\t-\\r ]"),
        ("\\d", "[\\d]"),
        ("[0123456789]", "[\\d]"),
        ("(0|1|2|3|4|5|6|7|8|9)", "([\\d])"),
        (r"This is (?!not\.)nothing\.", r"This is (?!not\.)nothing\."),
        (r"This is (?=not\.)nothing\.", r"This is (?=not\.)nothing\."),
        ("^start", "^start"),
        ("not^start", "not^start"),
        ("end$", "end$"),
        ("end$not", "end$not"),
        ("[a-z]", "[a-z]"),
        ("[^a-z]", "[^a-z]"),
        ("[^b]", "[^b]"),
        ("[a-zA-Z]", "[A-Za-z]"),
        ("[ababab]", "[ab]"),
        
        )
    for input, output in data:
        result = str(make_expression(input))
        if result != output:
            print repr(input), ":", repr(output), "!=", repr(result)
